import pandas as pd
import numpy as np
from tools.file import FileManager
import plotly.graph_objs as go
from .data_cleaner import DataCleaner
from market_research.scraper._visualization_scheme import BaseVisualizer
import requests
import plotly.io as pio
import plotly.express as px


class DataVisualizer(BaseVisualizer):
    def __init__(self, df:pd.DataFrame=None, maker="", output_folder_path="results"):
        
        pio.templates.default='ggplot2'
        self.colors = px.colors.qualitative.Plotly 
        self.markers = ['circle', 'x', 'square', 'star', 'diamond', 'pentagon', 'hexagon', 'cross', 'octagon', 'hexagon2']
        
        self.maker = maker
        super().__init__(output_folder_path = output_folder_path)
        FileManager.make_dir(output_folder_path)
        
        if df is not None:
            self.dc = DataCleaner(df)

    def price_map(self, return_data=False, return_fig=False):

        data = self.dc.get_price_df().copy()
        
        years = sorted(data['year'].unique(), reverse=True)
        series = data['series'].unique()
        
        color_map = {i: self.colors[i % len(self.colors)] for i in range(len(series))}
        marker_map = {year: self.markers[i] for i, year in enumerate(years)}
                 
        
        fig = go.Figure()
        for i, s in enumerate(series):
            data_series = data[data['series'] == s]
            year = data_series['year'].drop_duplicates().item()
        
            fig.add_trace(go.Scatter(
                x=data_series['size'],
                y=data_series['price'],
                mode='markers',
                marker=dict(
                    size=12,
                    color=color_map.get(i),  
                    symbol=marker_map.get(year),
                    opacity=0.8,  
                ),
                
                error_y=dict(
                    type='data',
                    symmetric=False,
                    arrayminus=[0]*len(data),
                    array=data_series['price_gap'],
                    color=color_map.get(i),  
                    thickness=1,
                    width=5
                ), 
                
                text=data_series['description'],  
                hoverinfo='text',  
                name=f'{s.upper()} ({year})',
                showlegend=True,
                visible=(year == years[0])
            ))
            
        
        year_dropdown = dict(
            buttons=[
                dict(
                    label=year,
                    method='update',
                    args=[
                        {
                            'visible': [year in trace.name for trace in fig.data]
                        }
                    ]
                ) for year in years
            ] +
            [
                dict(
                    label='All',
                    method='update',
                    args=[
                        {
                            'visible': [True] * len(fig.data)
                        }
                    ]
                )
            ]
        )

        ticks_below_3000 = list(range(0, 3001, 500))
        ticks_above_3000 = list(range(4000, max(int(data['price_original'].max()), int(data['price'].max())) + 2000, 1000))
        tickvals = ticks_below_3000 + ticks_above_3000

        fig.update_layout(
            updatemenus=[{
                'buttons': year_dropdown['buttons'],
                'direction': 'down',
                'showactive': True,
                'x': 0.1,  
                'y': 1.0,  
                'yanchor': 'bottom'  
            }],
            
            title='Price map',
            xaxis_title='Size (Inch)',
            yaxis_title='Current Price ($)',
            legend_title='Year',
            showlegend=True,
            template='simple_white',
            # ),
            yaxis=dict(
                range=[0, max(data['price_original'].max(),
                            data['price'].max()) + 1000],
                showgrid=True,
                tickmode='array',  
                tickvals=tickvals
            ),
            xaxis=dict(
                range=[30, 100],
                showgrid=True,
                tickmode='array',  
                tickvals=list(range(30, 100, 10))
            ),
            
            width=1000,
            height= 800,
            hovermode='closest',
            legend=dict(
                traceorder='reversed' 
            )
        )
        fig.write_html(self.output_folder/f"{self.maker}_pricemap.html")
        
        if return_fig:
            return fig
        else:
            fig.show()
        if return_data:
            data['series'] = data['series'].map(lambda x: f'{self.maker}_{x}')
            return data
        
        
            

    def power_consumption(self, return_data=False, return_fig=False):

        data = self.dc.get_power_concumption_df().copy()       
        years = sorted(data['year'].unique(), reverse=True)
        series = sorted(data['series'].unique())  
        
        color_map = {i: self.colors[i % len(self.colors)] for i in range(len(series))}
        marker_map = {year: self.markers[i] for i, year in enumerate(sorted(years, reverse=True))}
        
        fig = go.Figure()
        
        for i, s in enumerate(series):
            data_series = data[data['series'] == s]
            year = data_series['year'].drop_duplicates().item()
            price = data_series['price']
            
            
            fig.add_trace(go.Scatter(
                x=data_series['size'],
                y=data_series['maximum power consumption'],
                mode='markers',
                marker=dict(
                    size=12,  
                    color=color_map.get(i),  
                    symbol=marker_map.get(year),
                    opacity=0.8,  
                ), 
                name=f'{s.upper()} ({year})',
                hovertemplate='%{y}W (%{text}), $%{customdata[0]}<extra></extra>',  
                text=[s.upper()] * len(data_series),  
                customdata=np.stack([price], axis=-1),  
                showlegend=True,
                visible=(year == years[0])
            ))

        year_dropdown = dict(
            buttons=[
                dict(
                    label=year,
                    method='update',
                    args=[
                        {'visible': [year in trace.name for trace in fig.data]}
                    ]
                ) for year in years
            ] +
            [
                dict(
                    label='All',  
                    method='update',
                    args=[
                        {'visible': [True] * len(fig.data)}
                    ]
                )
            ]
        )

        fig.update_layout(
            updatemenus=[{
                'buttons': year_dropdown['buttons'],
                'direction': 'down',
                'showactive': True,
                'x': 0.1,  
                'y': 1.0,  
                'yanchor': 'bottom'  
            }],
            title='power consumption',
            legend_title='Year',
            showlegend=True,
            template='simple_white',
            yaxis=dict(
                title='W',
                range=[0, data['maximum power consumption'].max() + 100],  # y축 범위 조정
                showgrid=True,
                tickmode='linear',
                dtick=100  # y축의 간격을 100으로 설정
            ),
            xaxis=dict(
                range=[30, 100],
                showgrid=True,
                tickmode='array',  
                tickvals=list(range(30, 100, 10))
            ),

            width=1000,
            height=800,
            hovermode='closest',
            legend=dict(
                traceorder='reversed' 
            )
        )
        
        fig.write_html(self.output_folder/f"{self.maker}_power_cosumption.html")
        
        if return_fig:
            return fig
        else:
            fig.show()
        if return_data:
            data['series'] = data['series'].map(lambda x: f'{self.maker}_{x}')
            return data
        
        
        
    def heatmap_spec(self, col_selected:list=None, display_types:str=None, cmap="Blues", return_fig=False):
        if col_selected is None:
            try:
                file_path = "https://raw.githubusercontent.com/xikest/research_market_tv/main/json/col_heatmap.json"                
                response = requests.get(file_path)
                data = response.json()
                col_selected = data.get(self.maker)
            except Exception as e:
                file_path = "./json/col_heatmap.json"
                with open(file_path, 'r') as file:
                    import json
                    data = json.load(file)
                col_selected = data.get(self.maker)
                print(e)
        else:   
            if isinstance(col_selected, list):
                col_selected = col_selected
                col_selected =[c.lower() for c in col_selected]
            else:
                raise ValueError
        
        data_df = self.dc.get_df_cleaned().copy()
        available_columns = [col for col in col_selected if col in data_df.columns]
        missing_columns = [col for col in col_selected if col not in data_df.columns]
        if missing_columns:
            print("The following columns are missing from the DataFrame:")
            for col in missing_columns:
                print(col)

        data_df = data_df[available_columns]
        if display_types is not None:
            condition = data_df.index.get_level_values('display type').str.contains('|'.join(display_types), case=False, na=False)
            data_df = data_df[condition]
        data_df = data_df.mask(data_df == '-', 0)
        data_df = data_df.fillna(0)
        
        data_df = data_df[data_df.columns[::-1]]
        
        mask_data = data_df.reset_index().copy()
        data_df = data_df.map(lambda x: 1 if isinstance(x, str) else x)
        idx_names = data_df.index.names
        
        data_df['row_sum'] = data_df.sum(axis=1)
        data_df = data_df.reset_index().sort_values(by = idx_names + ['row_sum'], ascending=False).drop_duplicates(subset=idx_names).drop(['row_sum'], axis=1)
        mask_data = mask_data.loc[data_df.index, :]

        data_df = data_df.set_index(idx_names)
        data_df = data_df.sort_index(ascending=True)
        self.data_df = data_df
        heatmap_data = data_df.T
        
        mask_data = mask_data.replace(0, '').set_index(idx_names).sort_index(ascending=True).T
        
        x_labels = [f"{str(idx[1]).upper()} ({str(idx[0])})" for idx in heatmap_data.columns]
        
        years = sorted(heatmap_data.columns.get_level_values(0).unique(), reverse=True)        
        year_heatmapdf_dict = {}
        year_mask_dict = {}
        for year in years:
            year_heatmapdf = heatmap_data.xs(year, level=0, axis=1)
            year_maskdf = mask_data.xs(year, level=0, axis=1)
            year_heatmapdf_dict[year] = year_heatmapdf.loc[(year_heatmapdf != 0).any(axis=1)]
            year_mask_dict[year] = year_maskdf.loc[year_heatmapdf_dict[year].index, year_heatmapdf_dict[year].columns]
            
        initial_year = years[0]
        fig = go.Figure(data=go.Heatmap(
            z=year_heatmapdf_dict.get(initial_year).values,  
            x=[f"{str(idx[0]).upper()} ({initial_year})" for idx in year_heatmapdf_dict.get(initial_year).columns],  
            y=year_heatmapdf_dict.get(initial_year).index.str.capitalize(),   
            colorscale=cmap,         
            showscale=False,
            zmax=1,                  
            hoverinfo='text',        
            hovertemplate='<b>%{customdata}</b><extra></extra>',
            customdata=year_mask_dict.get(initial_year).values
            ))

        year_dropdown = {
            'buttons': [
                dict(
                    label=str(year),
                    method='update',
                    args=[{
                        'z': [year_heatmapdf_dict.get(year).values],
                        'x': [[f"{str(idx[0]).upper()} ({year})" for idx in year_heatmapdf_dict.get(year).columns]],
                        'y': [year_heatmapdf_dict.get(year).index.str.capitalize()],
                        'customdata':[year_mask_dict.get(year).values],
                    }]
                ) for year in years
            ] + [
                dict(
                    label='All',
                    method='update',
                    args=[{
                        'z': [heatmap_data.values],
                        'x': [x_labels],
                        'y': [heatmap_data.index.str.capitalize()],
                        'customdata':[mask_data.values],
                    }]
                )
            ]
        }

        fig.update_layout(
            title=f"{self.maker} spec",
            updatemenus=[dict(
                type="dropdown",
                buttons=year_dropdown['buttons'],
                direction="down",
                showactive=True,
                x=0.1,
                y=1.1,
                yanchor='bottom',
            )],
            xaxis=dict(title='',tickangle=90),
            yaxis=dict(title='',tickangle=0),
            width=800,  
            height=800  
        )

        fig.write_html(self.output_folder/f"{self.maker}_heatmap.html")

        if return_fig:
            return fig
        else:
            fig.show()  
           

    def plot_headertxt(self, data=None, return_fig=False):
        if data is not None and isinstance(data, pd.DataFrame):   
            data = data
        else:
            data = self.dc.get_price_df().copy()

        data['year'] = data['year'].astype('str')
        data = data[['year', 'series', 'text0']].drop_duplicates().dropna()
        data = data.groupby(['text0', 'year']).agg({
            'series': '/ '.join 
        }).reset_index()
        data = data.sort_values(by=['year', 'series'], ascending=[False, True]).reset_index(drop=True)
        marker_size = 16  

        fig = go.Figure()
        for _, group in data.groupby('year'):
            fig.add_trace(go.Scatter(
                x=[group['year'].iloc[0]] * len(group),  
                y=group['series'].str.upper(),  
                mode="markers+text",
                text=group['text0'].str.capitalize(), 
                textposition="middle right",  
                marker=dict(size=marker_size),  
                hovertext=group['series'].str.upper(), 
                hoverinfo="text",
                textfont=dict(size=16),  
                name=group['year'].iloc[0] 
            ))
            
        # 최대 연도 구하기
        min_year = data['year'].astype(int).min()
        dummy_years = [str(min_year + i + 1) for i in range(30)]

        # 더미 연도 추가
        for year in dummy_years:
            fig.add_trace(go.Scatter(
                x=[year] * 1, 
                y=[None],  
                mode='markers',  
                marker=dict(size=marker_size),  
                showlegend=False  
            ))
            
        # 레이아웃 업데이트
        fig.update_layout(
            xaxis=dict(
                showgrid=False,
                visible=False,
                categoryorder='array',  # 카테고리 순서 설정
                categoryarray=list(reversed(sorted(data['year'].unique())))  # 연도를 역순으로 설정
            ),
            yaxis=dict(showgrid=False),
            showlegend=True,
            height=600,
            width=800,
            plot_bgcolor='white',
            paper_bgcolor='white',
            font=dict(color='black'),
            legend=dict(
                orientation="h",    
                yanchor="top",      
                xanchor="center",   
                x=0.5,              
                y=0               
            )
        )
        
        if return_fig:
            return fig

        fig.show()


import time
from selenium.webdriver.common.keys import Keys
from bs4 import BeautifulSoup
import pandas as pd
from tqdm import tqdm
from market_research.scraper._scraper_scheme import Scraper
import requests

class Rurlsearcher(Scraper):
    def __init__(self, enable_headless=True):
        super().__init__(enable_headless=enable_headless)
        self.wait_time = 2
        

    def _get_search_src(self):
        file_path = "https://raw.githubusercontent.com/xikest/research_market_tv/main/json/rtings_keywords.json"
        response = requests.get(file_path)
        data = response.json()
        return data
        

    def get_keywords_for_search(self, maker: str = None, category: str = None):
        src_dict = self._get_search_src()['keywords']
        keywords = []

        for maker_key, category_dict in src_dict.items():
            if maker is None or maker_key == maker:
                for category_key, keyword_list in category_dict.items():
                    if category is None or category_key == category:
                        keywords.extend([f"{maker_key} {keyword}" for keyword in keyword_list])
        
        return keywords
    
    
    def get_urls_for_search(self, maker: str = None, category: str = None):
        src = self._get_search_src()
        src_dict = src['urls']
        urls = []
        for maker_key, category_dict in src_dict.items():
            if maker is None or maker_key == maker:
                for category_key, url_list in category_dict.items():
                    if category is None or category_key == category:
                        urls.extend(url_list)
        
        return urls

    def get_urls_from_web(self, keywords: list = None) -> list:
        urls_set = set()
        for keyword in tqdm(keywords):
            url = self._search_and_extract_url(search_query=keyword)
            if url is not None:
                urls_set.add(url)
        return list(urls_set)

    def get_urls_from_inputpath(self, intput_folder_path: str) -> list:
        self.set_data_path(intput_folder_path=intput_folder_path)

        urls = []
        file_list = self.intput_folder.glob('*')
        excel_files = [file for file in file_list if file.suffix in {'.xlsx', '.xls'}]
        for excel_file in excel_files:
            df = pd.read_excel(excel_file)
            urls.extend(df["urls"])
        return urls

    def _search_and_extract_url(self, search_query: str, base_url="https://www.rtings.com"):
        driver = self.web_driver.get_chrome()
        try:
            driver.get(base_url)
            search_input = driver.find_element("class name", "searchbar-input")
            search_input.send_keys(search_query)
            search_input.send_keys(Keys.RETURN)
            time.sleep(self.wait_time)
            page_source = driver.page_source
            soup = BeautifulSoup(page_source, 'html.parser')
            search_results = soup.find_all('div', class_='searchbar_results-main')
            extracted_urls = [result.find('a')['href'] for result in search_results]

            # 추출된 URL을 /로 분할하고, 검색 결과와 비교하여 일치하는 경우 반환
            for url in extracted_urls:
                if "tv/reviews" in url:
                    split_url = url.split('/')
                    if len(split_url) == 5:
                        return base_url + url
        finally:
            # 브라우저 종료
            driver.quit()
        return None



import numpy as np
import pandas as pd
class DataCleaner:
    """
    cleanup = DataCleaner(df)
    df = cleanup.get_df_cleaned()
    df_prices = cleanup.get_price_df()
    """
    def __init__(self, df):
        self.df = df.copy()
        self._preprocess_df()
        self._cleanup_columns()

    def _preprocess_df(self):
        self.df = self.df.sort_values(["year", "series", "size", "grade"], axis=0, ascending=False)
        def transform_text(x):
            if isinstance(x, str):
                x = x.replace("  ", " ")  # 두 개의 공백을 하나로 변경
                x = x.replace("™", "")  # ™ 제거
                x = x.replace("®", "")  # ® 제거
                x = x.replace("\n\n", "\n ")  # 이중 줄바꿈을 단일 줄바꿈으로 변경
                x = x.replace(" \n", "\n ")  # 공백 후 줄바꿈 처리
                x = x.strip()  # 앞뒤 공백 제거
                x = x.lower()  # 모두 소문자로 변경
            return x

        self.df = self.df.map(transform_text)
        self.df.columns = [transform_text(x) for x in self.df.columns]
        self.df = self.df.groupby('series').apply(lambda x: x.ffill().bfill()).reset_index(drop=True)
        self.df = self.df.sort_values(by=["series","display type"])
        self.df.columns = self.df.columns.str.strip()
        try:
            self.df = self.df.rename(columns={"rated power consumption": "maximum power consumption", "maximum power consumption": "maximum power consumption"})
            self.df['maximum power consumption'] = self.df['maximum power consumption'].astype(str).str.replace('w', '').str.strip()
        except: pass
        self.df = self.df.sort_values(by=['year', 'size', 'series'], ascending=[False, True, False])
        self.df['year'] = self.df['year'].astype('str')
        

    def get_price_df(self):
            df = self.df[
                ["year", "display type", "size", "series", "model", "price", "price_original", "price_gap", "description"]]
            df = df.sort_values(["price", "year", "display type", "series", "size", ], ascending=False)
            df = df.dropna(subset=['price'])  #
            df['price_gap'] = df['price_gap'].fillna(0)
            df['price_gap'] = df['price_gap'].map(lambda x: int(x))
            df = df.sort_values(by=['year',  'series'], ascending=True)
            df.loc[:, 'description'] = df.apply(lambda row: 
                        f"{row['description']}<br>release: ${row['price_original']}<br>price: ${row['price']} ({row['price_gap']}↓)"  
                                                    if row['price_gap'] != 0 else 
                                                    f"{row['description']}<br>price: ${row['price']}" , axis=1)
            return df
        
    def get_power_concumption_df(self):
        df = self.df.copy()
        try: 
            df['maximum power consumption'] = df['maximum power consumption'].replace('-', np.nan)
            df = df[df['maximum power consumption'].notna()]
            df['maximum power consumption'] = pd.to_numeric(df['maximum power consumption'], errors='coerce') 
            df = df.dropna(subset=['maximum power consumption']) 
        except:
            df['maximum power consumption'] = None  
        return df
        

    def _cleanup_columns(self):
        self.df = self.df.drop_duplicates()

    def get_df_cleaned(self):
        if self.df is not None:
            df = self.df
            df["display type"] = df["display type"].str.replace(r'display|4k ', '', regex=True)
            df = df.set_index(["year", "series", "display type"]).drop(
                ["model", "grade"], axis=1)
            df = df.fillna("-")
            
            return df


import pandas as pd
import plotly.graph_objects as go
import requests
import datetime
import pandas as pd
import re 



class Calendar:
    def __init__(self, API_KEY:str, CALENDAR_ID :str):
        self.API_KEY = API_KEY
        self.CALENDAR_ID=CALENDAR_ID
        events = self._get_all_events(API_KEY, CALENDAR_ID)
        self._data = self._create_dataframe(events)
        pass
    
    @property
    def data(self):  
        return self._data
        
    def _get_all_events(self, API_KEY, CALENDAR_ID):
        start_time = datetime.datetime(2000, 1, 1).isoformat() + 'Z'  
        url = f'https://www.googleapis.com/calendar/v3/calendars/{CALENDAR_ID}/events'
        params = {
            'key': API_KEY,
            'timeMin': start_time,
            'maxResults': 250,
            'singleEvents': True,
            'orderBy': 'startTime'
        }
        events = []  
        while True:
            response = requests.get(url, params=params)
            if response.status_code == 200:
                data = response.json()
                items = data.get('items', [])
                events.extend(items)  
                
                if 'nextPageToken' not in data:
                    break
                
                params['pageToken'] = data['nextPageToken']
            else:
                print(f'Error fetching events: {response.status_code} - {response.text}')
                break
        return events 


    def _create_dataframe(self, events):
        def extract_url(description):
            urls = re.findall(r'https?://[^\s<>"]+', description)  
            return ', '.join(urls) if urls else 'No URL'  
        event_list = []
        for event in events:
            event_data = {
                'summary': event.get('summary', 'No Title'),  # 이벤트 제목
                'start': event['start'].get('dateTime', event['start'].get('date')),  # 시작 시간
                'end': event['end'].get('dateTime', event['end'].get('date')),  # 종료 시간
                'description_url': extract_url(event.get('description', 'No Description')),  # URL 추출
                'location': event.get('location', 'No Location'),  # 위치
                'attendees': ', '.join([attendee['email'] for attendee in event.get('attendees', [])])  # 참석자
            }
            event_list.append(event_data)
        df = pd.DataFrame(event_list)
        return df


    def create_events_calendar(self, filter_year = 2024, month_interval=10, return_fig=False):

        def preprocess_data(df, month_interval = 10):        

            for i in range(len(df)):
                summary_text = df["summary"][i].split(":")[0]
                cleaned_summary = summary_text.replace("sony","").replace("'s'","").replace("inch'","").replace("-'"," ")[:10] + "..."
                df.at[i, "keywords_summary"] = cleaned_summary
        
            # 요일 및 색상 매핑
            day_mapping = {
                'Monday': 0,
                'Tuesday': 1,
                'Wednesday': 2,
                'Thursday': 3,
                'Friday': 4,
                'Saturday': 5,
                'Sunday': 6,
            }
    
            df['start'] = pd.to_datetime(df['start'])
            df['year'] = df['start'].dt.year
            df['month'] = df['start'].dt.to_period('M').dt.start_time
            df['day_of_week'] = df['start'].dt.day_name()
            df['day_index'] = df['day_of_week'].map(day_mapping)
            df['month_number'] = df['month'].dt.strftime('%B')
        
            month_interval = month_interval
            df['month_numeric'] = df['month'].dt.month * month_interval
            df['week_number'] = df['start'].dt.isocalendar().week - df['start'].dt.to_period('M').dt.start_time.dt.isocalendar().week + 1
            df['week_label'] = df['week_number'].astype(str) + 'w'
        
            return df
        
        month_colors = {
            1: 'red',
            2: 'green',
            3: 'blue',
            4: 'orange',
            5: 'purple',
            6: 'pink',
            7: 'brown',
            8: 'cyan',
            9: 'magenta',
            10: 'gold',
            11: 'lightblue',
            12: 'lime'
        }
        
        df = self._data
        df = preprocess_data(df, month_interval=month_interval)  
        df[df['year'] == filter_year]
        
        marker_size = 11
        
        fig = go.Figure()
        point_counter = {}

        for i in range(len(df)):
            key = (df['month_numeric'][i], df['week_number'][i])

            if key not in point_counter:
                point_counter[key] = 0
            else:
                point_counter[key] += month_interval/7

            y_offset = point_counter[key]
            month = df['month_numeric'][i] // month_interval
            marker_color = month_colors.get(month, 'black')

            # 이벤트에 대한 산점도 추가
            fig.add_trace(go.Scatter(
                x=[df['week_number'][i]],
                y=[df['month_numeric'][i] + y_offset],
                mode='markers+text',
                textposition='middle right',
                text=df["keywords_summary"][i],
                # textfont=dict(size=12),
                marker=dict(size=marker_size, color=marker_color),
                hovertemplate=f'{df["summary"][i]}<extra></extra>',
                showlegend=False,
            ))

            # 설명 링크 추가
            fig.add_trace(go.Scatter(
                x=[df['week_number'][i]],
                y=[df['month_numeric'][i] + y_offset],
                mode='text',
                text=f'<a href="{df["description_url"][i]}" target="_blank" style="color:blue;"> </a>',
                textfont=dict(size=15),
                textposition='middle center',
                hovertemplate=f'{df["summary"][i]}<extra></extra>',
                showlegend=False,
            ))


        fig.update_xaxes(
            tickvals=df['week_number'].unique(),
            ticktext=[f'{int(w)}w' for w in sorted(df['week_number'].unique(), reverse=True)],
            visible=False
        )

        fig.update_yaxes(
            tickmode='array',
            tickvals=[month_interval * i for i in range(1, 13)],
            ticktext=[
                'January', 'February', 'March', 'April', 'May',
                'June', 'July', 'August', 'September',
                'October', 'November', 'December'
            ],
            autorange='reversed',
            gridcolor='lightgray',
        )

        # 레이아웃 업데이트
        fig.update_layout(
            title='Events Calendar',
            width=600,
            height=800,
            font=dict(color='black'),
        )

        
        if return_fig:
            return fig
        else:
            fig.show()




import pandas as pd
import FinanceDataReader as fdr
from datetime import datetime
import yfinance as yf
import plotly.graph_objects as go
import plotly.subplots as sp
from datetime import datetime
import requests

class SONY_IR():
    def __init__(self):
        # self.tas = TextAnalysis()
        
        self.cleaning_words = [  #사전 필터링하는 단어
                            "half","fy2021","fy2020", "month","way", "input","earnings",
                            "forecast","please","terms","market","g","ns", "unit","assets",
                            "fy2022","levels","q","fy2023","numbers","result","units",
                            "factors","costs","ss","q1","part",'segment', 'quarter', 
                            'statements', 'business', 'question', 'yen', 'year', 'sony', 'results',
                            "end","q2","questioner",
                            "session","fy2024",
                            # "sale","plan","capacity","growth","demand",
                            "outlook","increase","investment",
                            "example","rate","flow","time","a2","a1","sfh","r","dtc", "statement",
                            "plan", "tax", "value", "term","capital", "growth","company","group", "service",
                            "risk","profit","minotrity"
                        ]     
        
        
        self.replacement_mapping = {  #사전에 교체하는 단어
                            "games": "game",
                            "plans": "plan",
                            "sensors": "sensor",
                            "changes": "change",
                            "risks": "risk",
                            "services": "service",
                            "margins": "margin",
                            "profits": "profit",
                            "wafers": "wafer",
                            "sizes": "size",
                            "customers": "customer",
                            "applications": "application",
                            "shortages": "shortage",
                            "expenses": "expense",
                            "sales":"sale",
                            "titles":"title",
                            "conditions":"condition",
                            "prices":"price",
                            "investments":"investment",
                            "rates":"rate",
                            "inventories":"inventory",
                            "uncertainties":"uncertainty",
                            "cameras":"camera",
                            "opportunities":"opportunity",
                            "volumes":"volume",
                            "costs":"cost",
                            "technologies":"technology",
                            "employees":"employee",
                            "companies":"company",
                            "creators":"creator",
                            "challenges":"challenge",
                            "businesses":"business",
                            "years":"year",
                            "electronics":"electronic",
                            "strategies":"strategy",
                            "electronics":"electronic",
                            "targets":"target",
                            "statements":"statement"
                        }
        pass

    def get_ir_script(self) -> dict:
        def check_url_exists(url: str) -> bool:
            try:
                response = requests.head(url)
                # 200 OK
                return response.status_code == 200
            except requests.RequestException:
                return False
        
        today = datetime.now()
        start_date = today.replace(year=today.year - 4)

        file_dict = {}
        
        base_url = "https://www.sony.com/en/SonyInfo/IR/library/presen/er/pdf/"
        years = range(start_date.year, today.year + 1)  
        quarters = range(1, 5)
        for year in years:
            for quarter in quarters:
                filename = f"{year%100}q{quarter}_qa"  
                url = f"{base_url}{filename}.pdf"
                if check_url_exists(url):  # URL 
                                file_dict[f"20{filename.upper()}"] = url

        base_url = "https://www.sony.com/en/SonyInfo/IR/library/presen/strategy/pdf"
        years = range(2020, today.year + 1)  

        for year in years:
            for quarter in quarters:
                filename = f"{year}/qa_E"
                url = f"{base_url}/{filename}.pdf"
                if check_url_exists(url):  # URL
                    file_dict[filename.replace('/', '_').upper()] = url

        df = pd.DataFrame(list(file_dict.items()), columns=['filename', 'url'])
        df.loc[:,'year'] = df.filename.map(lambda x: x[:4])
        df.loc[:,"category"] = df.filename.map(lambda x:  "Strategy" if x[-1] == "E" else "Earning")
        df.loc[:,"quarter"]= df[df["category"] == "Earning"].filename.map(lambda x: x[4:6].upper())
        df.loc[:,"quarter"] = df["quarter"].fillna("-")
        return df 
        
    def plot_financials_with_margin(self, ticker='SONY'):
        def format_value(value):

            if value >= 1_000_000_000_000:  # 1T 
                return f'{value/1_000_000_000_000:.1f}T'
            elif value >= 1_000_000_000:  # 1B 
                return f'{value/1_000_000_000:.1f}B'
            elif value >= 1_000_000:  # 1M 
                return f'{value/1_000_000:.1f}M'
            elif value >= 1_000:  # 1K 
                return f'{value/1_000:.1f}K'
            else:
                return str(value)
                
        
        stock = yf.Ticker(ticker)
        financials = stock.financials
        
        operating_income = financials.loc['Operating Income'].dropna()
        total_revenue = financials.loc['Total Revenue'].dropna()

        latest_quarter = sorted(operating_income.index, reverse=True)[0]
        cols = sorted(stock.quarterly_financials.columns)
        financials_q = stock.quarterly_financials[cols[cols.index(latest_quarter) + 1:]]
        
        operating_income_q = financials_q.loc['Operating Income'].cumsum()
        total_revenue_q = financials_q.loc['Total Revenue'].cumsum()

        operating_income = pd.concat([operating_income_q, operating_income], axis=0)
        total_revenue = pd.concat([total_revenue_q, total_revenue], axis=0)

        operating_income.index = pd.to_datetime(operating_income.index)
        total_revenue.index = pd.to_datetime(total_revenue.index)

        fig = go.Figure()
        
        fig.add_trace(go.Bar(
            x=total_revenue.index, 
            y=total_revenue.values, 
            name='Total Revenue',
            text=[format_value(val) for val in total_revenue.values],  
            textposition='auto',
            textfont=dict(color='black'), 
            hoverinfo='none'    
        ))
        
        fig.add_trace(go.Bar(
            x=operating_income.index, 
            y=operating_income.values, 
            name='Operating Income',
            text=[format_value(val) for val in operating_income.values], 
            textposition='outside',
            textfont=dict(color='white'), 
            hoverinfo='none'    
        ))

        fig.update_layout(
            title=f"{ticker} Financials (Revenue, Operating Income)",
            xaxis_title="Date",
            yaxis_title="Amount (in USD)",
            barmode='overlay', 
            legend=dict(
                x=0,  
                y=1,  
                traceorder='normal',
                orientation='v'  
            ),
            xaxis=dict(
                tickformat="%b %Y",  
                tickvals=total_revenue.index,  
                ticktext=total_revenue.index.strftime("%b %Y")  
            ),
        )
        return fig
    
    def plot_usd_jpy_and_japan_gdp(self):
        today = datetime.now()
        start_date = today.replace(year=today.year - 4)
        end_date = today

        start_date_str = start_date.strftime('%Y-%m-%d')
        end_date_str = end_date.strftime('%Y-%m-%d')

        usd_jpy = fdr.DataReader('FRED:DEXJPUS', start=start_date_str, end=end_date_str)  # USD/JPY 
        usd_mxn = fdr.DataReader('FRED:DEXMXUS', start=start_date_str, end=end_date_str)  # USD/MXN
        usd_myr = fdr.DataReader('FRED:DEXMAUS', start=start_date_str, end=end_date_str)  # USD/MYR 
        eur_usd = fdr.DataReader('FRED:DEXUSEU', start=start_date_str, end=end_date_str)  # EUR/USD 


        jpy_myr = usd_myr['DEXMAUS'] / usd_jpy['DEXJPUS']  # JPY/MYR
        jpy_mxn = usd_mxn['DEXMXUS'] / usd_jpy['DEXJPUS']  # JPY/MXN
        eur_jpy = eur_usd['DEXUSEU'] * usd_jpy['DEXJPUS']  # EUR/JPY


        fig = sp.make_subplots(
            rows=2, cols=1,  
            shared_xaxes=True,  
            subplot_titles=("USD/JPY and JPY/MXN", "EUR/JPY and JPY/MYR"),  
            specs=[[{"secondary_y": True}], [{"secondary_y": True}]] 
        )

        fig.add_trace(go.Scatter(x=usd_jpy.index, y=usd_jpy['DEXJPUS'], mode='lines', name='USD/JPY'), row=1, col=1)
        fig.add_trace(go.Scatter(x=jpy_mxn.index, y=jpy_mxn, mode='lines', name='JPY/MXN'), row=1, col=1, secondary_y=True)
        
        fig.add_trace(go.Scatter(x=eur_jpy.index, y=eur_jpy, mode='lines', name='EUR/JPY'), row=2, col=1)
        fig.add_trace(go.Scatter(x=jpy_myr.index, y=jpy_myr, mode='lines', name='JPY/MYR'), row=2, col=1, secondary_y=True)

        fig.update_layout(
            title="Exchange Rates",
            xaxis_title="",
            yaxis_title="",  # 첫 번째 y축 제목 설정
            template="plotly_dark",
            height=800  # 그래프 높이 설정
        )
        return fig



from bs4 import BeautifulSoup
import requests
import time
import pandas as pd
from tqdm import tqdm
import re
from selenium.webdriver.common.by import By
from tools.file import FileManager
from market_research.scraper._scraper_scheme import Scraper, Modeler, CustomException
from market_research.scraper.models.visualizer.data_visualizer import DataVisualizer



class ModelScraper_l(Scraper, Modeler, DataVisualizer):
    def __init__(self, enable_headless=True,
                 export_prefix="lge_model_info_web", intput_folder_path="input", output_folder_path="results",
                 verbose: bool = False, wait_time=2, demo_mode:bool=False):

        Scraper.__init__(self, enable_headless=enable_headless, export_prefix=export_prefix,
                         intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)

        self.tracking_log = verbose
        self.wait_time = wait_time
        self.file_manager = FileManager
        self.log_dir = "logs/lge/models"
        if self.tracking_log:
            FileManager.make_dir(self.log_dir)
            
        self._data = self._fetch_model_data(demo_mode=demo_mode)    
        DataVisualizer.__init__(self, df = self._data, maker='lge')
        pass
    
    @property
    def data(self):
        return self._data
    
    def _fetch_model_data(self, demo_mode:bool=False) -> pd.DataFrame:
        def find_urls() -> dict:
            url_set = set()
            url_series_set = self._get_series_urls()
            for url in tqdm(url_series_set):
                url_models_set = self._extract_models_from_series(url=url)
                url_set.update(url_models_set)
            url_dict = {idx: url for idx, url in enumerate(url_set)}
            print("total model:", len(url_dict))
            return url_dict
        
        def extract_sepcs(url_dict):
            dict_models = {}
            for key, url in tqdm(url_dict.items()):
                try:
                    dict_info = self._extract_model_details(url)
                    dict_models[key] = dict_info
                    dict_spec = self._extract_global_specs(url=url)
                    dict_models[key].update(dict_spec)
                except Exception as e:
                    if self.tracking_log:
                        print(f"fail to collect: {url}")
                        print(e)
            return dict_models
        
        def transform_format(dict_models, json_file_name: str) -> pd.DataFrame:
            df_models = pd.DataFrame.from_dict(dict_models).T
            df_models = df_models.drop(['Series', 'Size'], axis=1)
            df_models = df_models.dropna(subset=['price'])
            df_models.to_json(self.output_folder / json_file_name, orient='records', lines=True)
            return df_models
            
        if demo_mode:
            print("operating demo")
            df_models = pd.read_json('https://raw.githubusercontent.com/xikest/research_market_tv/main/json/l_scrape_model_data.json', orient='records', lines=True)
        else:
            print("collecting models")
            url_dict = find_urls()
            dict_models = extract_sepcs(url_dict)
            df_models = transform_format(dict_models, json_file_name="l_scrape_model_data.json")
            
        FileManager.df_to_excel(df_models.reset_index(), file_name=self.output_xlsx_name)
        return df_models
    
    @Scraper.try_loop(2)
    def _get_series_urls(self) -> set:

        def find_series_urls(url, prefix) -> set:
            url_series = set()
            url = url
            prefix = prefix
            step = 200
            try:
                driver = self.web_driver.get_chrome()
                driver.get(url=url)
                time.sleep(self.wait_time)
                scroll_distance_total = self.web_driver.get_scroll_distance_total()
                scroll_distance = 0

                while scroll_distance < scroll_distance_total:
                    for _ in range(2):
                        html = driver.page_source
                        soup = BeautifulSoup(html, 'html.parser')
                        elements = soup.find_all('a', class_="css-11xg6yi")
                        for element in elements:
                            url_series.add(prefix + element['href'].strip())
                        driver.execute_script(f"window.scrollBy(0, {step});")
                        time.sleep(self.wait_time)
                        scroll_distance += step
                return url_series
            except Exception as e:
                pass
            finally:
                driver.quit()
        
        url_series = find_series_urls( url = "https://www.lg.com/us/tvs", prefix = "https://www.lg.com/")
        print("The website scan has been completed.")
        print(f"total series: {len(url_series)}")
        return url_series

    @Scraper.try_loop(try_total=5)
    def _extract_models_from_series(self, url: str, prefix="https://www.lg.com/") -> set:
      
        def extract_model_url(elements)->set:
            url_models_set= set()
            for element in elements:
                try:
                    url = prefix + element['href']
                    url_models_set.add(url.strip())
                except Exception as e:
                    print(f"Getting series error ({e})")
                    pass
            return url_models_set
            
        url_models_set = set()
        CustomException(message=f"error_extract_models_from_series: {url}")
        try:
            soup = BeautifulSoup(requests.get(url).text, 'html.parser')
            elements = soup.find_all('a', class_='css-1a0ki8h')
            url_models_set = extract_model_url(elements)
        except CustomException as e:
            pass
        finally:
            pass
        return url_models_set
    
    @Scraper.try_loop(2)
    def _extract_model_details(self, url: str) -> dict:
        
        def extract_model(soup)->dict:
            
            label = soup.find('span', class_="MuiTypography-root MuiTypography-overline css-rrulv7").text.strip()
            model = label.split()[-1]
            return {"model": model}
        
        def extract_prices(soup)->dict:
            prices_dict = dict()
            try:
                price = soup.find('div', class_='MuiGrid-root MuiGrid-item css-8wacqv').text.strip()
                split_price = price.split('$')
                pattern = r'\d{1,3}(?:,\d{3})*(?:\.\d{2})?'
                prices = [re.search(pattern, part).group() for part in split_price if re.search(pattern, part)]
                if len(prices) == 3:
                    price_now = float(prices[0].replace(',', ''))
                    price_gap = round(float(prices[1].replace(',', '')), 1)
                    price_original = float(prices[2].replace(',', ''))
                    
                    prices_dict["price"] = price_now
                    prices_dict["price_gap"] = price_gap
                    prices_dict["price_original"] = price_original
                else:
                    price_now = float(split_price[-1].replace(',', '')) 
                    prices_dict["price"] = price_now
                    prices_dict["price_original"] = price_now
                    prices_dict["price_gap"] = 0.0
            except:
                prices_dict["price"] = float('nan')
                prices_dict["price_original"] = float('nan')
                prices_dict["price_gap"] = float('nan')
                
            return prices_dict
       
        def extract_description(soup)->dict:
            descriptions = [
                ('h2', 'MuiTypography-root MuiTypography-subtitle2 css-8oa1vg'),
                ('h1', 'MuiTypography-root MuiTypography-subtitle2 css-8oa1vg'),
                ('h1', 'MuiTypography-root MuiTypography-h5 css-vnteo9')
            ]
            for tag, class_name in descriptions:
                try:
                    description = soup.find(tag, class_=class_name).text.strip()
                    if description: break  
                except AttributeError:     
                    description =""
                
            return {"description": description}
        

        
        dict_info = {}
        if self.tracking_log: print(" Connecting to", url)
        response = requests.get(url)
        page_content = response.text
        soup = BeautifulSoup(page_content, 'html.parser')
        
        dict_info.update(extract_model(soup))
        dict_info.update(extract_prices(soup))
        dict_info.update(extract_description(soup))
        dict_info.update(ModelScraper_l.extract_info_from_model(dict_info.get("model")))
           
        if self.tracking_log:
            self._dir_model = f"{self.log_dir}/{dict_info.get('model')}"
            self.file_manager.make_dir(self._dir_model)
        return dict_info

    @Scraper.try_loop(try_total=10)
    def _extract_global_specs(self, url: str) -> dict:
      
        def set_driver(url):
            driver = self.web_driver.get_chrome()
            driver.get(url=url)
            time.sleep(self.wait_time)
            return driver
        

        def find_spec_tab(driver):
            try: 
                for id in range(5):
                    element_all_specs = driver.find_element(By.ID, f"simple-tab-{id}")
                    if element_all_specs.text.lower() == "all specs":
                        break
                self.web_driver.move_element_to_center(element_all_specs)
                time.sleep(self.wait_time)
                element_all_specs.click()
                time.sleep(self.wait_time)
            except:
                try: 
                    element_all_specs = driver.find_element(By.CLASS_NAME, "MuiTypography-root.MuiTypography-h6.MuiTypography-alignLeft.MuiLink-root.MuiLink-underlineAlways.css-kgbp8r")
                    self.web_driver.move_element_to_center(element_all_specs)
                    time.sleep(self.wait_time)
                except:
                    element_all_specs = driver.find_element(By.CLASS_NAME, 'MuiTypography-root.MuiTypography-h5.css-14uiqdv')
                    self.web_driver.move_element_to_center(element_all_specs)
                    time.sleep(self.wait_time)
                
                
            return None 
        

        def extract_spec_detail(driver) -> dict:
            dict_spec = {}
            spec_elements = driver.find_elements(By.CSS_SELECTOR, '.MuiBox-root.css-1nnt9ji')
            for spec_element in spec_elements:
                try: 
                    label = spec_element.find_element(By.CLASS_NAME,
                                                        'MuiTypography-root.MuiTypography-body3.css-11mszpq').text.strip()
                    content = spec_element.find_element(By.CLASS_NAME,
                                                        'MuiTypography-root.MuiTypography-body2.css-1yx8hz4').text.strip()
                except:
 
                    label = spec_element.find_element(By.CLASS_NAME,
                                                    'MuiTypography-root.MuiTypography-body3.css-byc8c0').text.strip()
                    content = spec_element.find_element(By.CLASS_NAME,
                                                    'MuiTypography-root.MuiTypography-body2.css-1yx8hz4').text.strip()

                
                

                if label != '':
                    
                    original_label = label
                    while label in dict_spec and dict_spec.get(label)!=content:
                        asterisk_count = label.count('*')
                        label = f"{original_label}{'*' * (asterisk_count + 1)}"
                    dict_spec[label] = content

            dict_spec.pop("*", None)            
            return dict_spec
       
        dict_spec = dict()
        CustomException(message=f"error_extract_global_specs: {url}")
        try:
            driver = set_driver(url)
            if self.tracking_log:
                stamp_today = self.file_manager.get_datetime_info(include_time=False)
 
                stamp_url = self.file_manager.get_name_from_url(url)
                driver.save_screenshot(f"./{self._dir_model}/{stamp_url}_0_model_{stamp_today}.png")                
            find_spec_tab(driver)   
            if self.tracking_log: driver.save_screenshot(f"./{self._dir_model}/{stamp_url}_1_element_all_specs_{stamp_today}.png")
            dict_spec = extract_spec_detail(driver)
            if self.tracking_log: print(f"Received information from {url}")
            return dict_spec
        except CustomException as e:
            pass
        finally:
            driver.quit()
            
    @staticmethod
    def extract_info_from_model(model: str)->dict:
        model = model.lower()  # 대소문자 구분 제거
        dict_info = {}

        # 연도 매핑
        year_mapping = {'oled':
                            {'1': "2021",
                            '2': "2022",
                            '3': "2023",
                            '4': "2024",
                            '5': "2025",
                            '6': "2026"},
                        'qned':
                            {'p': "2021",
                            'q': "2022",
                            'r': "2023",
                            'u': "2024"},
                        'u':{ 
                            'p': "2021",
                            'q': "2022",
                            'r': "2023",
                            't': "2024",}
                        }


        
        # "oled"가 포함된 모델 처리
        if "oled" in model:
            model = model[:-3].replace("oled", "")
            dict_info["grade"] = "oled"
            dict_info["year"] = model[-1]
            dict_info["series"] = model[-2:]
            dict_info["size"] = model[:-2]
            dict_info["year"] = year_mapping.get('oled').get(dict_info.get("year"), None)
            
        
        # "qned" 또는 "nano"가 포함된 모델 처리
        elif "qned" in model or "nano" in model:
            model = model[:-1] #마지막자리 삭제
            if "qned" in model:
                model_split = model.split("qned")
                dict_info["grade"] = "qned"
            else:
                model_split = model.split("nano")
                dict_info["grade"] = "nano"
            dict_info["size"] = model_split[0]
            model = model_split[-1]
            dict_info["year"] = model[-1]
            dict_info["series"] = model
            # dict_info["series"] = model[:-2]
            dict_info["year"] = year_mapping.get('qned').get(dict_info.get("year"), None)
            dict_info
        
        # "lx"가 포함된 모델 처리
        elif "lx" in model:
            model_split = model.split("lx")
            dict_info["grade"] = "flexble"
            dict_info["size"] = model_split[0]
            dict_info["series"] = model
            model = model_split[-1]
            dict_info["year"] = model[0]
            # dict_info["series"] = model[1]
            dict_info["year"] = year_mapping.get('oled').get(dict_info.get("year"), None)
        
        # "u"가 포함된 모델 처리
        elif "u" in model:
            model = model[:-3]
            model_split = model.split("u")
            dict_info["grade"] = "u"
            dict_info["size"] = model_split[0]
            model = model_split[-1]
            dict_info["year"] = model[0]
            dict_info["series"] = model
            dict_info["year"] =  year_mapping.get('u').get(dict_info.get("year"), None)
        return dict_info


from abc import ABC, abstractmethod
import pandas as pd
import platform
from pathlib import Path
from datetime import date
from tools.web import WebDriver


class Scraper(ABC):

    def __init__(self, enable_headless=True, export_prefix:str="scraper", intput_folder_path = None, output_folder_path=None):
        """
        enable_headless=True,
        export_prefix:str="scaper",
        intput_folder_path = "input",
        output_folder_path="results"
        """
        self.webdriver_path:str
        self.webdriver_path:str
        self.intput_folder:Path
        self.output_folder:Path
        self.output_xlsx_name = None

        self._set_webdriver_paths()
        self._initialize_data_paths(export_prefix=export_prefix, intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)

        self.web_driver = WebDriver(executable_path=self.webdriver_path, browser_path=self.browser_path, headless=enable_headless)
        self.wait_time = 1

    def _set_webdriver_paths(self):
        current_os = platform.system()
        self.webdriver_path = "./chromedriver/chromedriver.exe"
        self.browser_path = "./chrome/chrome.exe"

        if current_os == "Linux":
            self.webdriver_path = "/content/chromedriver/chromedriver"
            self.browser_path = "/content/chrome/chrome"

    def _initialize_data_paths(self,export_prefix:str=None, intput_folder_path:str=None, output_folder_path:str=None):
        if intput_folder_path is not None:
            self.intput_folder = Path(intput_folder_path)  # 폴더 이름을 지정
            if not self.intput_folder.exists():
                self.intput_folder.mkdir(parents=True, exist_ok=True)

        if output_folder_path is not None:
            self.output_folder = Path(output_folder_path)
            if not self.output_folder.exists():
                self.output_folder.mkdir(parents=True, exist_ok=True)
            if export_prefix is not None:
                self.output_xlsx_name = self.output_folder /  f"{export_prefix}{date.today().strftime('%Y-%m-%d')}.xlsx"
            else:
                self.output_xlsx_name = self.output_folder / f"{date.today().strftime('%Y-%m-%d')}.xlsx"
   
    @property
    def data(self):
        return self._data
              
    @staticmethod
    def try_loop(try_total):
        def decorator(func):
            def wrapper(*args, **kwargs):
                for cnt_try in range(try_total):
                    try:
                        return func(*args, **kwargs)  # 인스턴스 메서드 여부와 관계없이 호출
                    except Exception as e:
                        if cnt_try + 1 == try_total:
                            print(f"Error after {try_total} attempts: {e}")
                        continue  # 다음 시도로 넘어감
            return wrapper
        return decorator


class Modeler(ABC):

    @abstractmethod
    def _fetch_model_data(self, demo_mode: bool = False) -> pd.DataFrame:
        """
        Collect model information from URLs and return the data in the desired format.
        """
        pass

    @abstractmethod
    def _get_series_urls(self) -> set:
        """
        Get the series URLs by scrolling down the main page.
        """
        pass

    @abstractmethod
    def _extract_models_from_series(self, url: str) -> dict:
        """
        Extract all model URLs from a given series URL.
        """
        pass

    @abstractmethod
    def _extract_model_details(self, url: str) -> dict:
        """
        Extract model information (name, price, description) from a given model URL.
        """
        pass

    @abstractmethod
    def _extract_global_specs(self, url: str) -> dict:
        """
        Extract global specifications from a given model URL.
        """
        pass



class CustomException(Exception):
    def __init__(self, message:str=""):
        super().__init__(message) 




import pandas as pd
import seaborn as sns
import os
import plotly.graph_objects as go
from wordcloud import WordCloud
import matplotlib.pyplot as plt
import base64
import io
import wget
from nltk.probability import FreqDist
from nltk import pos_tag
from wordcloud import WordCloud
import matplotlib.pyplot as plt
from nltk.tokenize import word_tokenize
from nltk.corpus import stopwords
import nltk
import fitz  # PyMuPDF
from ._analysis_scheme import Analysis


class TextAnalysis(Analysis):
    def __init__(self,
                 export_prefix="text_", intput_folder_path="/content/input",  output_folder_path="results"):
        self.comments:list
        self.nouns:list
        self.df_word_freq:pd.DataFrame
        super().__init__(export_prefix=export_prefix,  intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)

        # self._set_datapack()
    def read_files_from_inputpath(self, docs_type="excel"):

        file_type = {"excel":['.xlsx', '.xls'],
                     "pdf":[".pdf"]

                     }
        file_list = self.intput_folder.glob('*')

        files = [file for file in file_list if file.suffix in file_type.get(docs_type)]
        return files


    def download_pdf(self, url:str):

        try:
            wget.download(url)
        except Exception as e:
            print(f"Error downloading {url}: {str(e)}")
        return 


    def download_pdfs(self, urls:list, mkdir=True) ->list:
        filepath_list= []
        for url in urls:
            try:
                filepath = os.path.join(self.intput_folder, os.path.basename(url))
                if mkdir and  not os.path.exists(filepath):
                    wget.download(url, filepath)
                else:
                    filepath = wget.download(url)
                filepath = os.path.join(os.getcwd(), filepath)
                filepath_list.append(filepath)
            except Exception as e:
                print(f"Error downloading {url}: {str(e)}")
        return filepath_list
        
        
    def set_comments(self, comments: list, cleaning_words: list = None) -> None:
        self.comments = comments
        self._prepare_nouns(cleaning_words)
        self._prepare_word_freq()
        return None

    def save_df_freq_as_excel(self, file_name: str = "your_excel_file_name"):
        df = self.df_word_freq
        df.to_excel(self.output_folder / (file_name + "_freq.xlsx"), index=False)

    def _set_datapack(self) -> None:
        nltk.download('punkt')
        nltk.download('stopwords')
        nltk.download('averaged_perceptron_tagger')
        return None
    
    def _prepare_nouns(self, cleaning_words) -> None:
        all_words = []
        for comment in self.comments:
            tokens = word_tokenize(comment)
            all_words.extend(tokens)
        stop_words = set(stopwords.words('english'))
        stop_words.update(cleaning_words)
        filtered_words = [word.lower() for word in all_words if
                          word.isalnum() and word.lower() not in stop_words]
        self._nouns = [word for (word, tag) in pos_tag(filtered_words) if tag.startswith('N')]
        return None
    def _prepare_word_freq(self) -> None:
        noun_counts = FreqDist(self._nouns)
        df_word_freq = pd.DataFrame(list(noun_counts.items()), columns=['Nouns', 'Frequency'])
        self.df_word_freq = df_word_freq.sort_values(by='Frequency', ascending=False)
        return None

    def plot_freq(self, num_dis: int = 10):
        try:
            df_word_freq=self.df_word_freq
            sns.set(style="white")
            top_words = df_word_freq.sort_values(by="Frequency", ascending=False).head(num_dis)
            plt.figure(figsize=(10, 4))
            sns.barplot(x='Nouns', y='Frequency', data=top_words)
            plt.title("Top 10 Words Frequency")
            plt.xticks(rotation=45)
            sns.despine()
            plt.show()

        except Exception as e:
            print(e)
            print("data not prepared")

    def plot_wordcloud(self, comment, cleaning_words=[]):
        
        self.set_comments([comment], cleaning_words)
                
        nouns = self._nouns

        width = 1000 
        height = 1000 

        try:
            wordcloud = WordCloud(width=width, height=height, background_color='white').generate(" ".join(nouns))

            img = io.BytesIO()
            wordcloud.to_image().save(img, format='PNG')
            img.seek(0)
            encoded_img = base64.b64encode(img.getvalue()).decode()

            fig = go.Figure()
            fig.add_layout_image(
                dict(
                    source=f'data:image/png;base64,{encoded_img}',
                    x=0, y=1,
                    xref="paper", yref="paper",
                    sizex=1, sizey=1,
                    xanchor="left", yanchor="top",
                    opacity=1,
                    layer="above"
                )
            )

            fig.update_layout(
                xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                width=width,
                height=height,
                margin=dict(l=0, r=0, t=0, b=0)  
            )

            # fig.show()
        except Exception as e:
            print(e)
            print("data not prepared")
        return fig
            
            

    def pdf_to_text(self, pdf_path):
        text = ''
        with fitz.open(pdf_path) as pdf_reader:
            num_pages = pdf_reader.page_count
            for page_num in range(num_pages):
                page = pdf_reader[page_num]
                text += page.get_text()
        
        return text

